from panelexpr.base.parser import *
from panelexpr.base.exprtree import ExpressionTree, DataProvider
from panelexpr.base.operator import TimeSeriesOperator, CrossSectionalOperator, DoubleSeriesOperator,global_operator_dict
from panelexpr.base import parser


__all__ = [
    "TimeSeriesOperator",
    "CrossSectionalOperator",
    "DoubleSeriesOperator",
    "eval",
    "register",
    "unregister"
]


def eval(formula, data, group_tag=None, time_tag=None, print_only=False):
    tree = parse(formula)
    ctx = {
        "kwargs": {
            "group_tag": group_tag,
            "time_tag": time_tag
        }
    }
    f_desc = ast2desc(tree, ctx)
    if print_only:
        print(f_desc)
        return None
    expr = ExpressionTree(f_desc, data_provider=DataProvider(data))
    return expr.values.rename()


def register(key, op_cls):
    ir_name = op_cls.__name__
    global_operator_dict.register(ir_name, op_cls)
    if issubclass(op_cls, TimeSeriesOperator):
        parser.register(key, ir_name, {"group_by": "group_tag"})
    elif issubclass(op_cls, CrossSectionalOperator):
        parser.register(key, ir_name, {"group_by": "time_tag"})
    else:
        parser.register(key, ir_name)


def unregister(key):
    ir_name = parser.FUN_MAPPING[key]
    global_operator_dict.unregister(ir_name)
    parser.unregister(key)




import numba
import numpy as np


@numba.jit
def corr(a1, a2):
    c = np.mean((a1-np.mean(a1)) * (a2-np.mean(a2)))
    v1 = np.var(a1)
    v2 = np.var(a2)
    return c / np.sqrt(v1 * v2)


@numba.jit
def cov(a1, a2):
    # 返回无偏结果
    c = np.sum((a1-np.mean(a1)) * (a2-np.mean(a2))) / (len(a1) - 1)
    return c


@numba.jit
def rolling_mean(a, lag):
    l = len(a)
    r = np.empty(l, dtype=np.float64)

    if l < lag:
        r[:] = np.nan
        return r

    f = 1
    s = 0
    for i in range(l):
        if not np.isnan(a[i]):
            s += a[i]
            if f < lag:
                f += 1
                r[i] = np.nan
            else:
                r[i] = s / lag
                s -= a[i - lag + 1]
        else:
            s = 0
            f = 1
            r[i] = np.nan

    return r


@numba.jit
def _rolling_var(a, lag):
    m = rolling_mean(a, lag)
    l = len(a)
    r = np.empty(l, dtype=np.float64)
    r[:lag-1] = np.nan

    if l < lag:
        return r

    for i in range(lag - 1, l):
        s = a[i - lag + 1: i + 1]
        r[i] = np.sum((s - m[i]) * (s - m[i]))
    return r


@numba.jit
def rolling_var(a, lag):
    return _rolling_var(a, lag) / (lag - 1)


@numba.jit
def rolling_std(a, lag):
    return np.sqrt(_rolling_var(a, lag) / (lag - 1))


@numba.jit
def _rolling_cov(a1, a2, lag):
    m1 = rolling_mean(a1, lag)
    m2 = rolling_mean(a2, lag)
    l = len(a1)
    r = np.empty(l, dtype=np.float64)
    r[:lag-1] = np.nan
    if l < lag:
        return r
    for i in range(lag - 1, l):
        s1 = a1[i - lag + 1: i + 1]
        s2 = a2[i - lag + 1: i + 1]
        r[i] = np.sum((s1 - m1[i]) * (s2 - m2[i]))
    return r


@numba.jit
def rolling_cov(a1, a2, lag):
    return _rolling_cov(a1, a2, lag) / (lag - 1)


@numba.jit
def rolling_corr(a1, a2, lag):
    c = _rolling_cov(a1, a2, lag)
    v1 = _rolling_var(a1, lag)
    v2 = _rolling_var(a2, lag)
    return c / np.sqrt(v1 * v2)


@numba.jit
def rolling(a, lag):
    l = len(a)
    r = np.empty(l, dtype=np.float64)
    r[:lag] = np.nan
    if l < lag:
        return r
    r[lag:] = a[:-lag]
    return r


@numba.jit
def ewma(a, span):
    # print(a)
    alpha = 2 / (span + 1)
    l = len(a)
    wa = np.empty(l, dtype=np.float64)

    p = np.nan
    f = 1
    wa[0] = p = a[0]
    for i in range(1, l):
        if np.isnan(a[i]):
            wa[i] = p
        else:
            if np.isnan(wa[i-1]):
                if np.isnan(p):
                    p = wa[i] = a[i]
                else:
                    wa[i] = p
            else:
                p = wa[i] = alpha * a[i] + (1 - alpha) * wa[i-1]
                f = 1
                # p = wa[i]


    return wa






from panelexpr.base.node import BaseNode
from panelexpr.base.operator import SlotOperator, ExtractOperator, drop_med, \
    global_operator_dict
import copy
from panelexpr.base.context import context_builder


def default_tree_builder(node, op_dict, context, data_provider, node_cls=BaseNode, **kwargs):
    if node is not None:
        try:
            name = node['name']
        except KeyError:
            name = None

        try:
            args = node['args']
        except KeyError:
            args = {}

        try:
            children = [default_tree_builder(child, op_dict, context, data_provider, node_cls, **kwargs) for child in node['children']]
        except KeyError:
            children = []
        op = op_dict[node['op']](**args)
        if isinstance(op, ExtractOperator):
            op.load(data_provider)
        return node_cls(name, op, args, None, children, context, **kwargs)
    # name = None, op = None, args_bak = None, parent = None, children = None, context = None
    return None


class ExpressionTree:
    def __init__(self, expression=None, data_provider=None,
                 op_dict=None, tree_builder=default_tree_builder,
                 expr_tree=None, cols=None):
        """

        :param expression: 表达式描述树
        :param tree_builder: 执行建树操作的函数
        :param op_dict: 描述符到操作符的映射
        :param expr_tree: （可选）已经转换为树状组织但尚未包装成为ExpressionTree的表达式
        """
        if op_dict is None:
            # op_dict = OperationDict()
            op_dict = global_operator_dict

        if data_provider is None:
            raise ValueError("No data is provided")

        self.data_provider = data_provider
        self._context = context_builder()
        if expression:
            expression = copy.deepcopy(expression)
            expression['name'] = "factor"
            self._expr_tree = tree_builder(expression, op_dict, self._context, self.data_provider)
        elif expr_tree:
            self._expr_tree = expr_tree
            self._expr_tree.parent = None
        else:
            raise ValueError("parameter 'expression' and 'expr_tree' can not be both None.")

        self.height = self._expr_tree.height
        self.count = self._expr_tree.count
        self._traverse_counter = 0
        self._post_order_list = None
        self.__values = None
        self.__slots = None
        self.__cols = cols

    @property
    def expr_tree(self):
        return self._expr_tree

    @property
    def values(self):
        if self.__values is None:
            self.__values = self._expr_tree.values()
            if self.__cols:
                self.__values = self.__values[self.__cols]
            self._expr_tree.reset()
        return self.__values

    def tolist(self):
        self._check_post_order_list()
        return self._post_order_list

    def to_dict(self):
        return self._expr_tree.to_dict()

    # @profile
    def cut_(self, item):
        """
        会修改自身的剪切方法
        :param item:
        :return:
        """
        self.reset()
        child = self._post_order_list[item]
        parent = child.parent
        if parent is None:

            branch = self.data_free_deepcopy()
            self._expr_tree = BaseNode(name=self._expr_tree.name, op=SlotOperator(), parent=None)
            self.reset()
            branch.reset()
            return self, branch

        child_rank = parent.find_child(child)
        parent.children[child_rank] = BaseNode(name=child.name, op=SlotOperator(), parent=parent)

        branch = ExpressionTree(data_provider=self.data_provider, expr_tree=child)

        # 清空过去保存的计算结果
        self.reset()
        branch.reset()

        # 返回剪切后的树和剪下来的计算树枝
        return self, branch

    def cut(self, item):
        tree = self.data_free_deepcopy()
        return tree.cut_(item)

    # @profile
    def graft_(self, branch):
        """
        注意，此操作可能导致data_provider不一致，因此在reset中增加统一dp的操作
        :param branch:
        :return:
        """
        self.__update_slots()
        if self.__slots:
            s = self.__slots[0]
        else:
            raise ValueError("There is no slot in the expression tree.")

        # self.reset()
        branch = branch.data_free_deepcopy()
        branch.reset()
        p = s.parent
        if p is None:
            name = self._expr_tree.name
            self._expr_tree = branch.expr_tree
            self._expr_tree.name = name
            self.reset()
            return self

        child_rank = p.find_child(s)
        child = branch.expr_tree
        p.children[child_rank] = child

        # 清空过去保存的计算结果
        self.reset()
        return self

    def graft(self, branch):
        tree = self.data_free_deepcopy()
        return tree.graft_(branch)

    def get_slots(self):
        self.__update_slots()
        return self.__slots

    def reset(self):
        self._expr_tree.parent = None
        self.__update_post_order_list()
        self.__update_slots()
        self.__values = None
        self.unify_data_provider()
        self._expr_tree.reset()
        self.data_provider.reset()

    # @profile
    def data_free_deepcopy(self):
        d = self.unlink_data()

        v = self.__values
        self.__values = None
        # s = asizeof.asizeof(self)
        # sh = self.data_provider.is_holding_data()

        cp = copy.deepcopy(self)
        # ch = cp.data_provider.is_holding_data()

        self.link_data(d)
        self.__values = v
        cp.link_data(d)
        return cp

    def unlink_data(self):
        d = self.data_provider.unlink_data()
        return d

    def link_data(self, d):
        self.data_provider.link_data(d)

    def unify_data_provider(self):
        self._expr_tree.unify_data_provider(self.data_provider)

    def __update_slots(self):
        self._check_post_order_list()
        self.__slots = [s for s in self._post_order_list if isinstance(s.op, SlotOperator)]

    def __update_post_order_list(self):
        self.__build_traverse_list()

    def __build_traverse_list(self):
        self._traverse_counter = 0
        self._post_order_list = []
        self.__post_order_traverse(self._expr_tree)

    def __post_order_traverse(self, expr_tree):
        if not expr_tree:
            return
        for child in expr_tree.children:
            self.__post_order_traverse(child)
        self._post_order_list.append(expr_tree)

    def _check_post_order_list(self):
        if not self._post_order_list:
            self.__build_traverse_list()

    # def __getitem__(self, item):
    #     return self.value()[item]

    def __repr__(self):
        return str(self.values)


class DataProvider:
    def __init__(self, data, use_copy=False):
        self._data = data
        self._use_copy = use_copy

    def __call__(self, cols = None):
        if cols is None:
            if self._use_copy:
                return self._data.copy()
            else:
                return self._data
        return self._data[cols]

    def unlink_data(self):
        data = self._data
        self._data = None
        return data

    def is_holding_data(self):
        return self._data is not None

    def link_data(self, data):
        self._data = data

    def reset(self):
        if not self._use_copy:
            drop_med(self._data)





from .operator import BaseOperator, ExtractOperator
from .operator import OpType
# from .operator import OperationDict
from panelexpr.base.operator import drop_med


class BaseNode:
    def __init__(self, name=None, op=None, args_bak=None, parent=None, children=None, context=None):
        self.__parent = parent
        if children is None:
            self.__children = []
        else:
            self.__children = children
        for child in self.__children:
            child.parent = self
        self._operator = op
        self._name = name
        self._kwargs_bak = args_bak
        self.__value = None
        self._height = None
        self._count = None
        self._context = context

    @property
    def height(self):
        if self._height is not None:
            return self._height
        if self.__children:
            self._height = max([t.height for t in self.__children]) + 1
        else:
            self._height = 0
        return self._height

    @property
    def count(self):
        if self._count is not None:
            return self._count
        if self.__children:
            self._count = sum([t.count for t in self.__children]) + 1
        else:
            self._count = 1
        return self._count

    @property
    def op(self):
        return self._operator

    @property
    def children(self):
        return self.__children

    @property
    def parent(self):
        return self.__parent

    @parent.setter
    def parent(self, p):
        self.__parent = p

    def reset_context(self, context):
        self._context = context
        for c in self.__children:
            c.reset_context(context)

    def unify_data_provider(self, data_provider):
        if isinstance(self._operator, ExtractOperator):
            self._operator.load(data_provider)
        for c in self.__children:
            c.unify_data_provider(data_provider)

    def find_child(self, child):
        i = 0
        for c in self.__children:
            if c == child:
                break
            i += 1
        return i

    # @profile
    def calculate(self):
        t = self._operator.operator_type

        if t == OpType.TERM:
            v = self._operator.exec(self._context)
        else:
            params = [node.calculate() for node in self.__children]
            v = self._operator.exec(self._context, *params)
        # drop_med(v[1], ignore=[v[0]])
        self.__value = v
        return self.__value

    # @profile
    def reset(self):
        self.__value = None
        self._height = None
        self._count = None
        for child in self.__children:
            child.reset()

    # @profile
    def values(self):
        if self.__value is None:
            self.calculate()
        if self._name is not None:
            r = self._context[self.__value[0]][self.__value[1]]
            # r = self.__value[1].rename(columns={self.__value[0]:self._name})
        else:
            # r = self.__value[1]
            r = self._context[self.__value[0]][self.__value[1]]

        return r

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, n):
        self._name = n

    def to_dict(self):
        d = {
            "op": self._operator.name(),

        }
        if self._kwargs_bak:
            d["args"] = self._kwargs_bak
        children = [c.to_dict() for c in self.__children]
        if children:
            d["children"] = children
        return d

    def __repr__(self):
        return str(self._operator)





import ast
import warnings


def parse(formula):
    tree = ast.parse(formula)
    tree = ast.fix_missing_locations(tree)
    return tree


def ast2desc(node, ctx):
    e = node.body[0].value
    return _ast2desc(e, ctx)


def _construct_unaryop(node, ctx):
    return {}


BINOP_MAPPING = {
    ast.Add: "Add",
    ast.Mult: "Mul",
    ast.Sub: "Sub",
    ast.Div: "Div"

}


def _construct_binop(node, ctx):
    d = {
        "op": BINOP_MAPPING[type(node.op)]
    }
    children = [
        _ast2desc(node.left, ctx),
        _ast2desc(node.right, ctx)
    ]
    d["children"] = children

    return d


def _keyword_parser(node, ctx):
    assert isinstance(node, ast.keyword)
    k = node.arg
    if isinstance(node.value, ast.Str):
        v = node.value.s
    else:
        v = None
    return k, v


FUN_MAPPING = {
    "shift": "Shift",
    "mmean": "MovingAverage",
    "ma": "MovingAverage",
    "mstd": "MovingStd",
    "mcov": "MovingCov",
    "mcorr": "MovingCorr",
    "ewm": "EWM",
    "ema": "EWM",
    "where": "Where",
    "less": "Less",
    "rank": "Rank",
    "mmax": "TimeSeriesMax",
    "mmin": "TimeSeriesMin"
}

NECESSARY_ARGS = {
    "shift": {"group_by": "group_tag"},
    "mmean": {"group_by": "group_tag"},
    "ma": {"group_by": "group_tag"},
    "mstd": {"group_by": "group_tag"},
    "mcov": {"group_by": "group_tag"},
    "mcorr": {"group_by": "group_tag"},
    "ewm": {"group_by": "group_tag"},
    "ema": {"group_by": "group_tag"},
    "where": None,
    "less": None,
    "rank": {"group_by": "time_tag"},
    "mmax": {"group_by": "group_tag"},
    "mmin": {"group_by": "group_tag"},
}


def _check_func(funcid):
    try:
        FUN_MAPPING[funcid]
    except KeyError:
        raise ValueError(f"no function with name {funcid}")


def _construct_call(node, ctx):
    _check_func(node.func.id)
    d = {
        "op": FUN_MAPPING[node.func.id]
    }
    children = [
        _ast2desc(n, ctx) for n in node.args
    ]
    kwargs = {}
    for k, v in [_keyword_parser(k, ctx) for k in node.keywords]:
        kwargs[k] = v
    # print(node.func.id, kwargs, ctx)
    if NECESSARY_ARGS[node.func.id]:
        nargs = NECESSARY_ARGS[node.func.id]
        for k, v in nargs.items():
            try:
                kwargs[k]
            except KeyError:
                kwargs[k] = ctx["kwargs"][v]

    d["args"] = kwargs
    d["children"] = children
    return d


def _construct_name(node, ctx):
    d = {
        "op": "Extract",
        "args": {"name": node.id}
    }
    return d


def _construct_num(node, ctx):
    d = {
        "op": "Literal",
        "args": {"value": node.n}
    }
    return d


CONSTRUCTOR = {
    ast.UnaryOp: _construct_unaryop,
    ast.BinOp: _construct_binop,
    ast.Call: _construct_call,
    ast.Name: _construct_name,
    ast.Num: _construct_num
}


def _ast2desc(node, ctx):
    try:
        # print(type(node))
        return CONSTRUCTOR[type(node)](node, ctx)
    except KeyError:
        print("KeyError with", type(node))
        return {}


def register(key, ir_name, necessary_args=None):
    try:
        t = FUN_MAPPING[key]
        warnings.warn(f"Function {key} is overridden.")
    except KeyError:
        pass
    finally:
        FUN_MAPPING[key] = ir_name
        NECESSARY_ARGS[key] = necessary_args


def unregister(key):
    del FUN_MAPPING[key]
    del NECESSARY_ARGS[key]



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="panelexpr", # Replace with your own username
    version="0.1.0a3",
    author="actcwlf",
    author_email="actcwlf@qq.com",
    description="A panel data evaluator.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="",
    license='MIT',
    packages=setuptools.find_packages(exclude=["_utils"]),
    classifiers=[
        'Development Status :: 3 - Alpha',
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'pandas>=0.20.0',
        'numpy>=1.14.0',
        'numba>=0.45.0',
        'numexpr>=2.7.0',

    ],
    python_requires='>=3.6',
)


from abc import abstractmethod, ABC, ABCMeta
import pandas as pd
import numpy as np
from enum import Enum
import panelexpr.boost.calc as bt
import warnings
"""
整体设计思路：
在子节点计算值保存在dataframe中后，同时向上返回列的名称和dataframe
装饰器根据这些信息提取出待计算的数据，送到相应exec函数中（将peeled_args传入），计算完毕后再以同样的方式向上层节点传递
"""

OpType = Enum("OperatorType", ["TERM", "UNARY", "BINARY", "TERNARY", "MULTI"])
label_recorder = 0
LABEL_PREFIX = "__ct_"


def get_label(op_str=""):
    global label_recorder
    label_recorder += 1
    return LABEL_PREFIX + op_str + "_" + str(label_recorder)


def is_label(s):
    return s[:len(LABEL_PREFIX)] == LABEL_PREFIX


def drop_med(data, ignore=None, inplace=True):
    if not isinstance(data, pd.DataFrame):
        return
    cols = list(data.columns)
    if ignore is not None:
        cols = set(cols) - set(ignore)
    med_col = [c for c in cols if is_label(c)]
    data.drop(columns=med_col, inplace=inplace)


def panel_wrapper(func):
    """
    确保输入输出均为带有标记的面板数据，适用于截面数据
    :param func:
    :return:
    """
    def wrapped(instance, *args):
        result = {}
        # 以靠前的dataframe优先作为保存结果的基础
        i = 0
        gid = []
        for arg in args:
            if isinstance(arg[1], pd.DataFrame):
                result = arg[1]
            else:
                gid.append(i)
            i += 1
        # 生成保存当前节点计算结果列的列名
        label = get_label(instance.__class__.__name__)
        # 将参数处理成便于计算的形式
        peeled_args = [arg[1][arg[0]] for arg in args]
        # 计算
        result[label] = func(instance, *peeled_args)

        # for i in gid:
        #     del args[i]
        # 按传递规则向上层返回
        return label, result

    return wrapped


class BasePanelMixin:
    @abstractmethod
    def extract(self, *args, **kwargs):
        pass

    def save(self, context, eval_result):
        # 生成保存当前节点计算结果列的列名
        label = get_label(self.__class__.__name__)
        context['data'][label] = eval_result
        # pass
        return 'data', label


class PanelMixin(BasePanelMixin):
    def extract(self, context, *args):
        operands = [context[k1][k2] for k1, k2 in args]
        return operands


class SingleGroupedPanelMixin(BasePanelMixin):
    """
    Single moving operator
    """
    def extract(self, context, key_col, groupby, key_num):
        """"""
        k1, k2 = key_col
        grouped_operand = context[k1].groupby(groupby)[k2]
        k1, k2 = key_num
        num_operand = context[k1][k2]
        return grouped_operand, num_operand


class CrossSectionGroupedPanelMixin(BasePanelMixin):
    """
    Single moving operator
    """
    def extract(self, context, key_col, groupby):
        """"""
        k1, k2 = key_col
        grouped_operand = context[k1].groupby(groupby)[k2]
        return grouped_operand


class DoubleGroupedPanelMixin(BasePanelMixin):
    """
    Single moving operator
    """
    def extract(self, context, key_col1, key_col2, groupby, key_num):
        """"""
        k1, k2 = key_col1
        grouped_operand1 = context[k1].groupby(groupby)[k2]
        k1, k2 = key_col2
        grouped_operand2 = context[k1].groupby(groupby)[k2]
        k1, k2 = key_num
        num_operand = context[k1][k2]
        return grouped_operand1, grouped_operand2, num_operand


def grouped_panel_wrapper(func):
    """
    确保输入输出均为带有标记的面板数据，适用于时间序列计算
    :param func:
    :return:
    """
    def wrapped(instance, *args):
        result = args[0][1]
        if not isinstance(result, pd.DataFrame):
            raise Exception("unmatched calculation")
        label = get_label(instance.__class__.__name__)

        if instance.sort:
            grouped = (
                args[0][1]
                .sort_values(
                    [instance.group_by, instance.sort_by],
                    ascending=instance.ascending
                )
                .groupby(instance.group_by)[args[0][0]]
            )
        else:
            grouped = args[0][1].groupby(instance.group_by)[args[0][0]]

        if instance.operator_type == OpType.UNARY:
            peeled_args = [grouped]
        else:
            peeled_args = [
                grouped,
                args[1][1][args[1][0]]
            ]
        # print(instance.__class__.__name__)
        # print(instance, *peeled_args)
        t = func(instance, *peeled_args)
        # print(t.shape)
        if isinstance(t.index, pd.MultiIndex):
            # print(t.index)
            # print(t.reset_index(drop=True))
            # print(t.droplevel(instance.group_by))
            try:
                result[label] = t.droplevel(instance.group_by)
            except:
                result[label] = t.reset_index(drop=True)
            # print(result)
        else:
            result[label] = t
        # del t
        # del peeled_args

        return label, result

    return wrapped


def grouped_panel_wrapper2(func):
    """
    确保输入输出均为带有标记的面板数据，适用于两个时间序列交叉计算
    目前已可以废弃
    :param func:
    :return:
    """
    def wrapped(instance, *args):
        result = args[0][1]
        if not isinstance(result, pd.DataFrame):
            raise Exception("unmatched calculation")
        label = get_label(instance.__class__.__name__)

        if instance.sort:
            grouped1 = (
                args[0][1]
                .sort_values(
                    [instance.group_by, instance.sort_by],
                    ascending=instance.ascending
                )
                .groupby(instance.group_by)[args[0][0]]
            )
            grouped2 = (
                args[1][1]
                .sort_values(
                    [instance.group_by, instance.sort_by],
                    ascending=instance.ascending
                )
                .groupby(instance.group_by)[args[1][0]]
            )
        else:
            grouped1 = args[0][1].groupby(instance.group_by)[args[0][0]]
            grouped2 = args[1][1].groupby(instance.group_by)[args[1][0]]

        t = pd.Series()
        for x, y in zip(grouped1, grouped2):
            t = t.append(func(instance, x[1], y[1], args[2][1][args[2][0]]))

        result[label] = t
        return label, result

    return wrapped


class BaseOperator:
    def __init__(self, op_type):
        self.__type = op_type

    @abstractmethod
    def exec(self, *args):
        pass

    @abstractmethod
    def eval(self, *args, **kwargs):
        pass

    @property
    def operator_type(self):
        return self.__type

    def name(self):
        return self.__class__.__name__[:-8]

    def __repr__(self):
        return self.__class__.__name__[:-8]


class BinaryOperator(BaseOperator, PanelMixin):
    def __init__(self):
        super().__init__(OpType.BINARY)

    def eval(self, lope, rope):
        pass

    def exec(self, context, *args):
        """
        extracting operands from context, executing operation and saving the result to the context
        :param context:
        :param args: keys of operands in context
        :return: key of result in context
        """
        operands = self.extract(context, *args)
        eval_result = self.eval(*operands)
        return self.save(context, eval_result)


class TernaryOperator(BaseOperator, ABC):
    def __init__(self, *args):
        super().__init__(OpType.TERNARY)

    def exec(self, *args):
        pass


class MultiOperator(BaseOperator, ABC):
    def __init__(self, *args):
        super().__init__(OpType.MULTI)

    def exec(self, *args):
        pass


class SingleSeriesOperator(BaseOperator, SingleGroupedPanelMixin, ABC):
    def __init__(self, op_type, group_by, sort=False, sort_by=None, ascending=True):
        super().__init__(op_type)
        self.group_by = group_by
        self.sort = sort
        self.sort_by = sort_by
        self.ascending = ascending

    def exec(self, context, *args):
        grouped_operand, num_operand = self.extract(context, args[0], self.group_by, args[1])
        eval_result = pd.Series()
        for k, operand in grouped_operand:
            eval_result = eval_result.append(self.eval(operand, num_operand))
        return self.save(context, eval_result)


class TimeSeriesOperator(SingleSeriesOperator, ABC):
    def __init__(self, group_by, sort=False, sort_by=None, ascending=True):
        super().__init__(OpType.BINARY, group_by, sort, sort_by, ascending)


class CrossSectionalOperator(BaseOperator, CrossSectionGroupedPanelMixin, ABC):
    def __init__(self, group_by):
        super().__init__(OpType.UNARY)
        self.group_by = group_by

    def exec(self, context, *args):
        grouped_operand = self.extract(context, args[0], self.group_by)
        eval_result = pd.Series()
        for k, operand in grouped_operand:
            eval_result = eval_result.append(self.eval(operand))
        return self.save(context, eval_result)


class DoubleSeriesOperator(BaseOperator, DoubleGroupedPanelMixin, ABC):
    def __init__(self, group_by, sort=False, sort_by=None, ascending=True):
        super().__init__(OpType.TERNARY)
        self.group_by = group_by
        self.sort = sort
        self.sort_by = sort_by
        self.ascending = ascending

    def exec(self, context, *args):
        grouped_operand1, grouped_operand2, num_operand = self.extract(context, args[0], args[1], self.group_by, args[2])
        eval_result = pd.Series()
        for x, y in zip(grouped_operand1, grouped_operand2):
            eval_result = eval_result.append(self.eval(x[1],y[1], num_operand))
        return self.save(context, eval_result)


class ExtractOperator(BaseOperator, ABC):
    """
    从数据源中提取数据
    """
    def __init__(self, name):
        super().__init__(OpType.TERM)
        self.__name = name
        self.__data_provider = None

    def load(self, data_provider):
        self.__data_provider = data_provider

    def exec(self, context):
        context['data'] = self.__data_provider()
        return 'data', self.__name

    def __repr__(self):
        return self.__class__.__name__[:-8] + ": '" + self.__name + "'"


class SlotOperator(BaseOperator, ABC):
    def __init__(self):
        super().__init__(OpType.TERM)

    def exec(self, context):
        label = get_label(self.__class__.__name__)
        context['literal'][label] = np.nan
        return 'literal', label


class LiteralOperator(BaseOperator):
    def __init__(self, value):
        super().__init__(OpType.TERM)
        self.__value = value

    def exec(self, context):
        label = get_label(self.__class__.__name__)
        context['literal'][label] = self.__value
        return 'literal', label

    def __repr__(self):
        if isinstance(self.__value, str):
            value = "'" + self.__value + "'"
        else:
            value = str(self.__value)

        return self.__class__.__name__[:-8] + ": " + value


class AddOperator(BinaryOperator):
    def eval(self, lope, rope):
        return pd.eval('lope + rope')


class SubOperator(BinaryOperator):
    def eval(self, lope, rope):
        return pd.eval('lope - rope')


class MulOperator(BinaryOperator):
    def eval(self, lope, rope):
        return pd.eval('lope * rope')


class DivOperator(BinaryOperator):
    def eval(self, lope, rope):
        return pd.eval('lope / rope')


class LessOperator(BinaryOperator):
    def eval(self, lope, rope):
        return lope < rope


class ShiftOperator(TimeSeriesOperator):
    def eval(self, series, lag):
        return series.shift(lag)


class TimeSeriesMaxOperator(TimeSeriesOperator):
    def eval(self, series, lag):
        return series.rolling(lag).max()


class TimeSeriesMinOperator(TimeSeriesOperator):
    def eval(self, series, lag):
        return series.rolling(lag).min()


class EWMOperator(TimeSeriesOperator):
    def eval(self, series: pd.Series, lag):
        return series.ewm(span=lag, min_periods=lag-1).mean()


class MovingAverageOperator(TimeSeriesOperator):
    def eval(self, series, window):
        s = pd.Series(bt.rolling_mean(series.to_numpy(), window), index=series.index)
        return s


class MovingStdOperator(TimeSeriesOperator):
    def eval(self, series, window):
        return pd.Series(bt.rolling_std(series.to_numpy(), window), index=series.index)


class WhereOperator(TernaryOperator):
    @panel_wrapper
    def exec(self, con, val1, val2):
        return np.where(con, val1, val2)


class MovingCovOperator(DoubleSeriesOperator):
    def eval(self, s1, s2, lag):
        v = pd.Series(bt.rolling_cov(s1.to_numpy(), s2.to_numpy(), lag), index=s1.index)
        return v


class MovingCorrOperator(DoubleSeriesOperator):
    def eval(self, s1, s2, lag):
        v = pd.Series(bt.rolling_corr(s1.to_numpy(), s2.to_numpy(), lag), index=s1.index)
        return v


class RankOperator(CrossSectionalOperator):
    def eval(self, cross):
        return cross.rank()


class OperatorDict:
    def __init__(self):
        self.mapping = {}

    def register(self, key, op_cls):
        try:
            t = self.mapping[key]
            warnings.warn(f"Operator {key} is overridden.")
        except KeyError:
            pass
        finally:
            self.mapping[key] = op_cls

    def unregister(self, key):
        del self.mapping[key]

    def __getitem__(self, key):
        try:
            return self.mapping[key]
        except KeyError:
            raise KeyError(f"Undefined operator {key}.")


global_operator_dict = OperatorDict()

global_operator_dict.register("Add", AddOperator)
global_operator_dict.register("Sub", SubOperator)
global_operator_dict.register("Mul", MulOperator)
global_operator_dict.register("Div", DivOperator)
global_operator_dict.register("Extract", ExtractOperator)
global_operator_dict.register("Literal", LiteralOperator)
global_operator_dict.register("Shift", ShiftOperator)
global_operator_dict.register("MovingAverage", MovingAverageOperator)
global_operator_dict.register("MovingStd", MovingStdOperator)
global_operator_dict.register("MovingCov", MovingCovOperator)
global_operator_dict.register("MovingCorr", MovingCorrOperator)
global_operator_dict.register("EWM", EWMOperator)
global_operator_dict.register("Where", WhereOperator)
global_operator_dict.register("Less", LessOperator)
global_operator_dict.register("Rank", RankOperator)
global_operator_dict.register("TimeSeriesMax", TimeSeriesMaxOperator)
global_operator_dict.register("TimeSeriesMin", TimeSeriesMinOperator)